@@ -2,7 +2,7 @@
  *      ipvsadm - IP Virtual Server ADMinistration program
  *                for IPVS NetFilter Module in kernel 2.4
  *
- *      Version: $Id: ipvsadm.c,v 1.19 2002/07/11 14:26:44 wensong Exp $
+ *      Version: $Id: ipvsadm.c,v 1.19.2.1 2002/11/14 10:05:26 wensong Exp $
  *
  *      Authors: Wensong Zhang <wensong@linuxvirtualserver.org>
  *               Peter Kese <peter.kese@ijs.si>
@@ -119,7 +119,7 @@
 #include "libipvs/libipvs.h"
 
 #define IPVSADM_VERSION_NO		"v" VERSION
-#define IPVSADM_VERSION_DATE		"2002/07/09"
+#define IPVSADM_VERSION_DATE		"2002/11/12"
 #define IPVSADM_VERSION			IPVSADM_VERSION_NO " " IPVSADM_VERSION_DATE
 
 #define MINIMUM_IPVS_VERSION_MAJOR      1
@@ -167,7 +167,8 @@ static const char* cmdnames[] = {
 #define OPT_DAEMON	0x00800
 #define OPT_STATS	0x01000
 #define OPT_RATE	0x02000
-#define NUMBER_OF_OPT	14
+#define OPT_SORT	0x04000
+#define NUMBER_OF_OPT	15
 
 static const char* optnames[] = {
 	"numeric",
@@ -184,6 +185,7 @@ static const char* optnames[] = {
 	"daemon",
 	"stats",
 	"rate",
+	"sort",
 };
 
 /*
@@ -196,22 +198,22 @@ static const char* optnames[] = {
  */
 static const char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] =
 {
-	/*     -n   -c   svc  -s   -p   -M   -r   fwd  -w   -mc  -to  dmn  -st -rt */
-/*INSERT*/    {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 'x', 'x', 'x', 'x'},
-/*ADD*/       {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*EDIT*/      {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*DEL*/       {'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*FLUSH*/     {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*LIST*/      {' ', '1', '1', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '1', '1', ' ', ' '},
-/*ADD-SERVER*/{'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', 'x', 'x', 'x', 'x', 'x'},
-/*DEL-SERVER*/{'x', 'x', '+', 'x', 'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*EDIT-SRV*/  {'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', 'x', 'x', 'x', 'x', 'x'},
-/*TIMEOUT*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*START-D*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', ' ', 'x', 'x', 'x', 'x'},
-/*STOP-D*/    {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*RESTORE*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*SAVE*/      {' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*ZERO*/      {'x', 'x', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+	/*     -n   -c   svc  -s   -p   -M   -r   fwd  -w   -mc  -to  dmn  -st  -rt  srt */
+/*INSERT*/    {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x'},
+/*ADD*/       {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*EDIT*/      {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*DEL*/       {'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*FLUSH*/     {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*LIST*/      {' ', '1', '1', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '1', '1', ' ', ' ', ' '},
+/*ADD-SERVER*/{'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*DEL-SERVER*/{'x', 'x', '+', 'x', 'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*EDIT-SRV*/  {'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*TIMEOUT*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*START-D*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', ' ', 'x', 'x', 'x', 'x', 'x'},
+/*STOP-D*/    {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*RESTORE*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*SAVE*/      {' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*ZERO*/      {'x', 'x', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
 };
 
 /* printing format flags */
@@ -220,6 +222,7 @@ static const char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] =
 #define FMT_RULE	0x0002
 #define FMT_STATS	0x0004
 #define FMT_RATE	0x0008
+#define FMT_SORT	0x0010
 
 #define SERVICE_NONE    0x0
 #define SERVICE_ADDR    0x1
@@ -245,7 +248,7 @@ static char * port_to_anyname(unsigned short port, unsigned short proto);
 static char * addrport_to_anyname(struct in_addr *addr, unsigned short port,
 				  unsigned short proto, unsigned int format);
 static int parse_service(char *buf, u_int16_t proto,
-		  u_int32_t *addr, u_int16_t *port);
+			 u_int32_t *addr, u_int16_t *port);
 static int parse_netmask(char *buf, u_int32_t *addr);
 static int parse_timeout(char *buf, int min, int max);
 static unsigned int parse_fwmark(char *buf);
@@ -351,6 +354,7 @@ parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		{"daemon", '\0', POPT_ARG_NONE, NULL, '6'},
 		{"stats", '\0', POPT_ARG_NONE, NULL, '7'},
 		{"rate", '\0', POPT_ARG_NONE, NULL, '8'},
+		{"sort", '\0', POPT_ARG_NONE, NULL, '9'},
 		{NULL, 0, 0, NULL, 0}
 	};
 
@@ -514,6 +518,10 @@ parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 			set_option(options, OPT_RATE);
 			*format |= FMT_RATE;
 			break;
+		case '9':
+			set_option(options, OPT_SORT);
+			*format |= FMT_SORT;
+			break;
 		default:
 			fail(2, "invalid option");
 		}
@@ -561,44 +569,45 @@ parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 	int c, parse;
 	const char *optstring = "AEDCZSRaedlLhvt:u:f:s:M:p::w:r:gmicn";
 	const struct option long_options[] =
-	{
-		{"add-service", 0, 0, 'A'},
-		{"edit-service", 0, 0, 'E'},
-		{"delete-service", 0, 0, 'D'},
-		{"clear", 0, 0, 'C'},
-		{"list", 0, 0, 'L'},
-		{"zero", 0, 0, 'Z'},
-		{"add-server", 0, 0, 'a'},
-		{"edit-server", 0, 0, 'e'},
-		{"delete-server", 0, 0, 'd'},
-		{"help", 0, 0, 'h'},
-		{"version", 0, 0, 'v'},
-		{"save", 0, 0, 'S'},
-		{"restore", 0, 0, 'R'},
-		{"set", 1, 0, '4'},
-		{"start-daemon", 1, 0, '1'},
-		{"stop-daemon", 0, 0, '2'},
-		{"tcp-service", 1, 0, 't'},
-		{"udp-service", 1, 0, 'u'},
-		{"fwmark-service", 1, 0, 'f'},
-		{"scheduler", 1, 0, 's'},
-		{"persistent", 2, 0, 'p'},
-		{"real-server", 1, 0, 'r'},
-		{"masquerading", 0, 0, 'm'},
-		{"netmask", 1, 0, 'M'},
-		{"ipip", 0, 0, 'i'},
-		{"gatewaying", 0, 0, 'g'},
-		{"weight", 1, 0, 'w'},
-		{"numeric", 0, 0, 'n'},
-		{"connection", 0, 0, 'c'},
-		{"mcast-interface", 1, 0, '3'},
-		{"timeout", 0, 0, '5'},
-		{"daemon", 0, 0, '6'},
-		{"stats", 0, 0, '7'},
-		{"rate", 0, 0, '8'},
-		{"help", 0, 0, 'h'},
-		{0, 0, 0, 0}
-	};
+		{
+			{"add-service", 0, 0, 'A'},
+			{"edit-service", 0, 0, 'E'},
+			{"delete-service", 0, 0, 'D'},
+			{"clear", 0, 0, 'C'},
+			{"list", 0, 0, 'L'},
+			{"zero", 0, 0, 'Z'},
+			{"add-server", 0, 0, 'a'},
+			{"edit-server", 0, 0, 'e'},
+			{"delete-server", 0, 0, 'd'},
+			{"help", 0, 0, 'h'},
+			{"version", 0, 0, 'v'},
+			{"save", 0, 0, 'S'},
+			{"restore", 0, 0, 'R'},
+			{"set", 1, 0, '4'},
+			{"start-daemon", 1, 0, '1'},
+			{"stop-daemon", 0, 0, '2'},
+			{"tcp-service", 1, 0, 't'},
+			{"udp-service", 1, 0, 'u'},
+			{"fwmark-service", 1, 0, 'f'},
+			{"scheduler", 1, 0, 's'},
+			{"persistent", 2, 0, 'p'},
+			{"real-server", 1, 0, 'r'},
+			{"masquerading", 0, 0, 'm'},
+			{"netmask", 1, 0, 'M'},
+			{"ipip", 0, 0, 'i'},
+			{"gatewaying", 0, 0, 'g'},
+			{"weight", 1, 0, 'w'},
+			{"numeric", 0, 0, 'n'},
+			{"connection", 0, 0, 'c'},
+			{"mcast-interface", 1, 0, '3'},
+			{"timeout", 0, 0, '5'},
+			{"daemon", 0, 0, '6'},
+			{"stats", 0, 0, '7'},
+			{"rate", 0, 0, '8'},
+			{"sort", 0, 0, '9'},
+			{"help", 0, 0, 'h'},
+			{0, 0, 0, 0}
+		};
 
 	extern char *optarg;
 	extern int optind;
@@ -782,6 +791,10 @@ parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 			set_option(options, OPT_RATE);
 			*format |= FMT_RATE;
 			break;
+		case '9':
+			set_option(options, OPT_SORT);
+			*format |= FMT_SORT;
+			break;
 		default:
 			fail(2, "invalid option");
 		}
@@ -1414,6 +1427,10 @@ print_service(struct ip_vs_service_user *svc, unsigned int format)
 	if (!(d = ipvs_get_dests(svc)))
 		exit(1);
 
+	if (format & FMT_SORT) {
+		ipvs_sort_dests(d, ipvs_cmp_dests);
+	}
+
 	if (svc->fwmark) {
 		if (format & FMT_RULE)
 			sprintf(svc_name, "-f %d", svc->fwmark);
@@ -1550,6 +1567,10 @@ static void list_all(unsigned int format)
 		exit(1);
 	}
 
+	if (format & FMT_SORT) {
+		ipvs_sort_services(get, ipvs_cmp_services);
+	}
+
 	print_title(format);
 	for (i=0; i<get->num_services; i++)
 		print_service(&get->entrytable[i], format);
